"""Module containing TablePrinter class for creating and formatting tables"""
from textwrap import fill

import colorful
from terminaltables import AsciiTable

colorful.use_style("solarized")


MAX_COLUMN_WIDTH = 80

DEFAULT_COLUMN_COLORS = [colorful.yellow, colorful.blue, colorful.green]


class TablePrinter(object):
    """Wrapper class around the terminaltables' :py:class:`AsciiTable`"""

    def __init__(self, use_colors=True):
        """Initialize an empty table"""
        self.table = AsciiTable([])
        self.use_colors = use_colors

    def __str__(self):
        """Overwrite in order to enable direct printing of the object"""
        return self.table.table

    def build_from_info(self, table_info):
        """Given a dictionary of information about a table describing an object,
        overwrite :py:attr:`.table` with the new content. An example
        `table_info` would look like this:

        .. code-block:: json

            {
                "rows": [
                    ["0", "foo"],
                    ["1", "bar"]
                ],
                "columns": ["Index", "Data"]
            }

        :param table_info: a dictionary of information including the column
            names and rows
        :type table_info: ``dict``
        """
        final_rows = []
        for row in table_info["rows"]:
            new_row = []
            for column_index, cell in enumerate(row):
                new_cell = fill(cell, MAX_COLUMN_WIDTH)
                if self.use_colors:
                    new_cell = _style_cell(new_cell, column_index)
                new_row.append(new_cell)
            final_rows.append(new_row)
        headers = table_info["columns"]
        if self.use_colors:
            headers = _style_headers(headers)
        self.table = AsciiTable([headers] + final_rows)


def _style_headers(headers):
    """Apply styles to the headers of a table.

    :param headers: list of header strings for styling
    :type headers: ``list`` of ``str``

    :returns: list of header strings that have been styled
    :rtype: ``list`` of ``str``
    """
    return [
        colorful.bold & DEFAULT_COLUMN_COLORS[i] | header
        for i, header in enumerate(headers)
    ]


def _style_cell(cell, column_index):
    """Apply styles to a cell of a table, taking care to account for newlines.

    :param cell: the cell string contents to style
    :type cell: ``str``
    :param column_index: the index of the column the cell belongs to
    :type column_index: ``int``

    :returns: a styled cell string
    :rtype: ``str``
    """
    segments = cell.split("\n")
    if len(segments) == 1:
        return DEFAULT_COLUMN_COLORS[column_index] | cell

    # Need to apply the coloring to each segment separately to prevent
    # terminaltables from incorrectly coloring the table border characters
    segments = [DEFAULT_COLUMN_COLORS[column_index] | segment for segment in segments]

    # Concat the segments back together (colorful objects don't support `join`)
    new_cell = ""
    for segment in segments:
        new_cell += segment
    return new_cell



"""Module containing the core :py:class:`Delver` object."""
from collections import namedtuple, OrderedDict

import six
from six.moves import input as six_input

from delver.exceptions import DelverInputError, ObjectHandlerInputValidationError
from delver.handlers import DictHandler, GenericClassHandler, ListHandler, ValueHandler
from delver.table import TablePrinter

DEFAULT_DIVIDER = "-" * 79


class Delver(object):
    """Object used for exploring arbitrary python objects.

    This class creates the runtime, manages state/user input, and coordinates
    the various object handler classes.

    The primary method is :py:meth:`run`, which starts the runtime.
    """

    object_handler_classes = [
        ListHandler,
        DictHandler,
        GenericClassHandler,
        ValueHandler,
    ]

    def __init__(self, target, verbose=False, use_colors=False):
        """Initialize the relevant instance variables.

        This includes the object handlers as well as those variables
        representing the runtime state.

        :param target: the object to delve into
        :param verbose: whether or not to allow object handlers to be verbose
        :type verbose: ``bool``
        """
        # The initial object which is set to enable returning to original state
        self._root_object = target

        # A named tuple to store information about the path, including
        # a 'path' attribute which is a list of object accessor strings
        # (like '["foo"]', '[0]'), and a 'previous' attribute which is just
        # a pointer to the previous object
        self._path = namedtuple("path", ["path", "previous"])
        self._path.path = ["root"]
        self._path.previous = []

        # The indicator for whether or not to continue program flow
        self._continue_running = False

        # An ordered mapping of possible user input commands to the function to
        # use if that input is selected
        self._basic_input_map = OrderedDict()
        self._basic_input_map["u"] = self._navigate_up
        self._basic_input_map["q"] = self._quit

        # Whether or not to allow object handlers to be verbose
        self._verbose = verbose

        self._use_colors = use_colors

        # The instantiated object handler classes from the class attribute
        # :py:attr:`.object_handler_classes`
        self._object_handlers = self._initialize_handlers()

    def _build_prompt(self, index_descriptor=None):
        """Create the user input prompt based on the possible commands.

        Builds the prompt from the :py:attr:`._basic_input_map`, taking into
        account the need for an index, as given by *index_descriptor*. An
        example would be '[<Key Index>, u, q] --> ' if *index_descriptor* was
        'Key Index'. If *index_descriptor is `None`, then the prompt would
        simply be '[u, q] --> '.

        :param index_descriptor: the description to use for the index input,
            e.g. 'Key Index'
        :type index_descriptor: ``str``

        :returns: the prompt to be given to the user for input
        :rtype: ``str``
        """
        basic_inputs = ", ".join(self._basic_input_map.keys())
        if index_descriptor is not None:
            prompt = "[<{}>, {}] --> ".format(index_descriptor, basic_inputs)
        else:
            prompt = "[{}] --> ".format(basic_inputs)
        return prompt

    def run(self):
        """Initialize the delver runtime.

        Handle initiating tables, coordinating the control of the currently
        in-scope object, and handle user input.

        The control flow will continue until :py:attr:`._continue_running` is
        set to ``False`` or a keyboard interrupt is detected.
        """
        target = self._root_object
        self._continue_running = True
        try:
            while self._continue_running:
                table = TablePrinter(use_colors=self._use_colors)
                for object_handler in self._object_handlers:
                    if object_handler.check_applies(target):
                        object_info = object_handler.describe(target)
                        table.build_from_info(object_info)
                        prompt = self._build_prompt(
                            index_descriptor=object_info.get("index_descriptor")
                        )
                        break

                self.print_table(table, description=object_info.get("description"))
                inp = six_input(str(prompt))
                target = self._handle_input(inp, target, object_handler)
        except (KeyboardInterrupt, EOFError):
            self.print_message("\nBye.")

    def print_table(self, table, description=None):
        """Prints the table and other supporting information for the view.

        This includes the divider to separate the view from other the previous
        one, the current path in the object, and a description of the current
        object.

        :param table: the table instance which contains the ascii table and
            cells
        :type table: :py:class:`.TablePrinter`
        :param description: an optional object description
        :type description: ``str``
        """
        view = []
        view.append(DEFAULT_DIVIDER)
        if len(self._path.path) > 0:
            view.append("At path: {}".format("".join(self._path.path)))
        if description is not None:
            view.append(description)
        view.append(str(table))
        six.print_("\n".join(view))

    def print_message(self, message):
        """Prints a generic message, generally either a warning or error"""
        six.print_(message)

    def _initialize_handlers(self):
        """Initialize handlers based on :py:attr:`._object_handler_classes`.

        :returns: list of object handler instances
        :rtype: ``list`` of :py:class:`.BaseObjectHandler`-derivced instances
        """
        instantiated_object_handlers = []
        for handler_class in self.object_handler_classes:
            instantiated_object_handlers.append(handler_class(verbose=self._verbose))
        return instantiated_object_handlers

    def _navigate_up(self, target):
        """Move to the previous parent object, making use of :py:attr:`._path`.

        :param target: the object representing the current location

        :returns: the parent object based on :py:attr:`._path`
        """
        if len(self._path.previous) == 0:
            self.print_message("Can't go up a level; we're at the top")
        else:
            target = self._path.previous.pop()
            self._path.path = self._path.path[:-1]
        return target

    def _quit(self, target):
        """End the primary program flow."""
        self.print_message("Bye.")
        self._continue_running = False

    def _handle_input(self, inp, target, object_handler):
        """Coordinate performing actions based on the user input.

        Checks the *inp* against the basic functions first, then attempts to
        use the *object_handler*'s own input handler.

        :param inp: the user-given input
        :type inp: ``str``
        :param target: the current object
        :param object_handler: the object handler which was applied to *target*
            for delving
        :type object_handler: :py:class:`.BaseObjectHandler`

        :returns: a (potentially) new object based on how the input is handled
        """
        if self._basic_input_map.get(inp) is not None:
            # Run the associated basic input handler function
            target = self._basic_input_map[inp](target)
        else:
            new_path = None
            try:
                old_target = target
                target, new_path = object_handler.handle_input(target, inp)
            except ObjectHandlerInputValidationError as err:
                self.print_message(err.msg)
            except DelverInputError:
                msg = ("Invalid command; please specify" " one of ['<{}>', {}]").format(
                    object_handler.index_descriptor,
                    ", ".join(self._basic_input_map.keys()),
                )
                self.print_message(msg)
            if new_path is not None:
                self._path.previous.append(old_target)
                self._path.path.append(six.text_type(new_path))
        return target


def run(target, **kwargs):
    """Initialize and begin execution of a :py:class:`Delver` object"""
    delver = Delver(target, **kwargs)
    delver.run()



"""General use exceptions for the delver"""


class ObjectHandlerInputValidationError(Exception):
    """Used when an object handler invalidates input"""

    def __init__(self, msg):
        self.msg = msg


class DelverInputError(Exception):
    """Used when an invalid argument is given to the delver"""

    def __init__(self, msg):
        self.msg = msg



#!/usr/bin/env python
# Copyright (c) 2020, Narrative Science
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# <COPYRIGHT HOLDER> BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.#
"""Delver is a helpful utility for stepping though nested objects"""
from __future__ import absolute_import, print_function

import argparse
import importlib
import json
import logging
import sys

import six

from delver.core import run


def _get_cli_args():
    """Parses and returns arguments passed from CLI.

    :return: parsed command line arguments
    :rtype: :py:class:`argparse.Namespace`
    """
    parser = argparse.ArgumentParser(
        description="Delve into JSON payloads from the command line."
    )
    parser.add_argument(
        "payload", type=argparse.FileType("r"), help="payload file to load"
    )
    parser.add_argument(
        "--transform-func",
        type=six.text_type,
        help=(
            "the module containing the optional json transform function, "
            'formatted like: "transform_module:transform_func". Note that the '
            "transform_func should take only a json-loaded dictionary as "
            "input and should return just the transformed dictionary."
        ),
    )
    return parser.parse_args()


def main():
    """System entry point."""
    my_args = _get_cli_args()

    payload_str = my_args.payload.read()
    try:
        payload = json.loads(payload_str)
    except ValueError:
        sys.exit('"{}" does not contain valid JSON.'.format(my_args.payload.name))

    if my_args.transform_func is not None:
        try:
            # We need to try and import the transform func and use that to
            # convert the payload before exploring
            transform_module_str, transform_func_str = my_args.transform_func.split(":")
            transform_module = importlib.import_module(transform_module_str)
            transform_func = getattr(transform_module, transform_func_str)
            payload = transform_func(payload)
        except ImportError:
            sys.exit(
                "Unable to import module `{}`. Please adjust the "
                "transform-func argument and try again.".format(transform_module_str)
            )
        except AttributeError:
            sys.exit(
                "Unable to import transform function `{}` "
                "from module `{}`. Please adjust the transform-func argument "
                "and try again.".format(transform_func_str, transform_module_str)
            )
        except Exception as error:
            logging.error(
                "Performing the transform function failed with the following "
                "error, please adjust the transform-func argument and try "
                "again:\n {}.".format(error)
            )
            raise error

    del payload_str
    my_args.payload.close()
    run(payload, use_colors=True, verbose=False)


if __name__ == "__main__":
    main()



"""Module containing object handler classes used by :py:class:`.Delver`"""

import six

from delver.exceptions import DelverInputError, ObjectHandlerInputValidationError


class BaseObjectHandler(object):
    """Base Object Handler class from which other handlers should inherit"""

    #: The object type associated with this handler
    handle_type = None

    #: The description of the input identifier integer, e.g. 'Key Index'
    index_descriptor = None

    #: The format string to wrap the object's path for display in the Delver
    path_modifier = "{}"

    def __init__(self, verbose=False):
        """Instantiate the necessary instance arguments"""
        self._encountered_pointer_map = {}
        self.verbose = verbose

    def check_applies(self, target):
        """Determine whether or not this object handler applies to this object
        based on it's type matching :py:attr:`.handle_type`.

        :returns: if the object is applicable to this handler
        :rtype: ``bool``
        """
        return isinstance(target, self.handle_type)

    def describe(self, target):
        """Create a table of information describing the contents of *target*.
        Must be implemented by child classes.

        :param target: the object to build the table information for

        :returns: a dictionary of information about *target* to be printed in a
            table
        :rtype: ``dict`` containing keys `columns`, `rows`, and optionally
            `description`
        """
        raise NotImplementedError("Child object handlers must overwrite this method")

    def handle_input(self, target, inp):
        """Validate the input and retrieve the relevant attribute of *target*
        based on the *inp*.

        :param target: the object to access
        :param inp: the user input
        :type inp: ``str``

        :returns: the appropriate attribute of *target*

        :raises :py:class:`.ObjectHandlerInputValidationError`: if the input
            is invalid for the given handler
        """
        inp = self._validate_input_for_obj(target, inp)
        return self._object_accessor(target, inp)

    def _object_accessor(self, target, inp):
        """The method called to actually perform the attribute accessing,
        which varies based on :py:attr:`.handle_type`. Must be implemented
        by child classes.

        :returns: the relevant attribute of *target*
        """
        raise NotImplementedError("Child object handlers must overwrite this method")

    def _add_property_map(self, target, index_prop_map):
        """Add an entry to this handler's encountered property map. This
        is used to keep track of the attributes associated with the object
        as well as their corresponding accesor indices.

        This mechanism relies on the fact that the id of the target
        remains the same, as do its attributes.

        :param target: the object to add the property map of
        :param index_prop_map: a mapping of the integers associated with
            *target*'s attributes to the respective attribute
        :type index_prop_map: ``dict``
        """
        self._encountered_pointer_map[id(target)] = index_prop_map

    def _validate_input_for_obj(self, target, inp):
        """Determine whether or not the given raw *inp* is valid for *target*
        as well as adjust *inp*'s type according to what is appropriate for
        this handler. This method can be overridden for more granular control.

        :returns: the input to be used for accessing the object's properties

        :raises :py:class:`.ObjectHandlerInputValidationError`: if the input
            is invalid for the given handler
        """
        try:
            inp = int(inp)
        except ValueError:
            raise DelverInputError("Invalid command")
        if inp not in self._encountered_pointer_map[id(target)].keys():
            raise ObjectHandlerInputValidationError("Invalid Index")
        return inp


class ListHandler(BaseObjectHandler):
    """Object handler for lists"""

    handle_type = list
    index_descriptor = "int"
    path_modifier = "[{}]"

    def describe(self, target):
        """Create the table info for the given list

        :param target: the list to build the table information for

        :returns: a dictionary of information detailing the elements in
            *target* and a high-level description
        :rtype: ``dict``
        """
        rows = []
        if len(target) == 0:
            column_names = ["Data"]
            index_descriptor = None
            rows.append([six.text_type("")])
        else:
            column_names = ["Idx", "Data"]
            index_descriptor = self.index_descriptor
            for i, value in enumerate(target):
                description = _get_object_description(value)
                rows.append([six.text_type(i), six.text_type(description)])
        object_info = {
            "columns": column_names,
            "rows": rows,
            "description": "List (length {})".format(len(target)),
            "index_descriptor": index_descriptor,
        }
        return object_info

    def _object_accessor(self, target, inp):
        """Get the *inp*-th element of *target* and the path accessor string"""
        return target[inp], self.path_modifier.format(inp)

    def _validate_input_for_obj(self, target, inp):
        """Make sure the *inp* is not greater than the length of *target*"""
        msg = None
        inp = int(inp)
        if inp >= len(target):
            msg = "Invalid index `{}`".format(inp)
            raise ObjectHandlerInputValidationError(msg)
        return inp


class DictHandler(BaseObjectHandler):
    """Object handler for dicts"""

    handle_type = dict
    index_descriptor = "key index"
    path_modifier = "[{}]"

    def describe(self, target):
        """Create the table info for the given dict

        :param target: the dict to describe

        :returns: dictionary of information about the keys and values in
            *target*
        :rtype: ``dict``
        """
        keys = sorted([(str(k), k) for k in target.keys()], key=lambda x: x[0])
        index_prop_map = {i: k for i, k in enumerate(keys)}
        self._add_property_map(target, index_prop_map)
        rows = []
        if len(keys) == 0:
            column_names = ["Data"]
            index_descriptor = None
            rows.append([six.text_type("")])
        else:
            column_names = ["Idx", "Key", "Data"]
            index_descriptor = self.index_descriptor
            for i, key_pair in enumerate(keys):
                value = target[key_pair[1]]
                description = _get_object_description(value)
                rows.append(
                    [
                        six.text_type(i),
                        six.text_type(key_pair[0]),
                        six.text_type(description),
                    ]
                )
        object_info = {
            "columns": column_names,
            "rows": rows,
            "index_descriptor": index_descriptor,
            "description": "Dict (length {})".format(len(target)),
        }
        return object_info

    def _object_accessor(self, target, inp):
        """Access the field in *target* associated with the key *inp* according
        to the :py:attr:`._encountered_pointer_map`.

        :param target: the dictionary to access
        :type target: ``dict``
        :param inp: the user-inputted key index associated with the desired
            field to access
        :type inp: ``int``

        :return: a ``tuple`` containing the desired field from *target* and the
            string accessor to describe the key needed to access that field
            directly
        """
        accessor_pair = self._encountered_pointer_map[id(target)][inp]
        if isinstance(accessor_pair[1], six.string_types):
            path_addition = '"{}"'.format(accessor_pair[1])
        else:
            path_addition = "{}".format(accessor_pair[1])
        return (target[accessor_pair[1]], self.path_modifier.format(path_addition))


class GenericClassHandler(BaseObjectHandler):
    """Object handler for any generic object, with functionality for
    optionally delving into the private methods/attributes of objects.
    """

    index_descriptor = "attr index"
    path_modifier = ".{}"
    _builtin_types = (int, float, bool, str, type(None))

    def check_applies(self, target):
        """Use this handler if *target* is not one of the basic built-ins"""
        return not isinstance(target, self._builtin_types)

    def describe(self, target):
        """Find the properties and methods of *target* and return information
        that will be used to build a table. Uses :py:attr:`.verbose` to
        determine whether or not to include private attributes/methods.

        :param target: the targetect to build information for

        :returns: a dictionary of information about *target*'s properties and
            methods
        :rtype: ``dict``
        """
        object_info = {}
        props = [prop for prop in _dir(target)]
        if not self.verbose:
            props = [prop for prop in props if not prop.startswith("_")]
        index_prop_map = {i: k for i, k in enumerate(props)}
        self._add_property_map(target, index_prop_map)
        if len(props) == 0:
            object_info["columns"] = ["Attribute"]
            rows = [[six.text_type(target)]]
            object_info["has_children"] = False
            object_info["index_descriptor"] = None
        else:
            object_info["columns"] = ["Idx", "Attribute", "Data"]
            object_info["index_descriptor"] = self.index_descriptor
            rows = []
            for i, prop in enumerate(props):
                attr = getattr(target, prop)
                description = _get_object_description(attr)
                rows.append(
                    [six.text_type(i), six.text_type(prop), six.text_type(description)]
                )
        object_info["rows"] = rows
        return object_info

    def _object_accessor(self, target, inp):
        """Retrieve the relevant property from *target* based on the property
        index *inp*.

        :param target: the object to retrieve the property from
        :param inp: the integer corresponding to the desired property from the
            :py:attr:`._encountered_pointer_map`
        :type inp: ``int``

        :returns: the desired property
        """
        attr_name = self._encountered_pointer_map[id(target)][inp]
        return getattr(target, attr_name), self.path_modifier.format(attr_name)


class ValueHandler(BaseObjectHandler):
    """Basic handler for single values"""

    has_children = False

    def check_applies(self, target):
        """Since this handler is always valid, simply return `True`"""
        return True

    def describe(self, target):
        """Build simple table info for a single value

        :param target: the value to describe

        :returns: a dictionary of information about the value
        :rtype: ``dict``
        """
        object_info = {}
        object_info["columns"] = ["Value"]
        if target is None:
            description = six.text_type("None")
        else:
            description = six.text_type(target)
        object_info["rows"] = [[description]]
        return object_info


def _get_object_description(target):
    """Return a string describing the *target*"""
    if isinstance(target, list):
        data = "<list, length {}>".format(len(target))
    elif isinstance(target, dict):
        data = "<dict, length {}>".format(len(target))
    else:
        data = target
    return data


def _dir(target):
    """Wrapper function to enable testing of builtin functions"""
    return dir(target)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['delver']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['colorful>=0.4.0', 'six>=1.10.0', 'terminaltables>=3.1.0']

entry_points = \
{'console_scripts': ['delve = delver.delve:main']}

setup(name='pydelver',
      version='0.0.6',
      description='Helpful tool for stepping through nested objects',
      author='Alex Sippel',
      author_email='asippel@narrativescience.com',
      url='https://github.com/NarrativeScience/delver',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6,<4',
     )
